﻿using UnityEngine;
using System.Collections.Generic;


/// <summary>
/// ボス
/// </summary>
public class Boss : MonoBehaviour {
	private const int BULLET_MAX = 500;                 // プール最大数
	private const int WAY_COUNT = 8;                    // WAY数
	private const float WAY_ANGLE = 360f / WAY_COUNT;   // 全方位なので360度をWAY数で割る
	private const float SHOT_SPAN = 12f / 60f;          // 連射間隔（sec.）※1秒間に12発
	private const float SHOT_SPEED = 4f;                // 弾速（m/sec.）
	private const float RAPID_ANGLE = 5.5f;             // 連射毎の回転角（degree）
	private static readonly Quaternion WAY_ROT = Quaternion.Euler(0f, WAY_ANGLE, 0f);   // WAY間の回転値

	
	public GameObject explosion = null;	// 爆発オブジェクト
	public int endurance = 350;         // 耐久値（シリアライズ値が優先されるのでデフォルト値）

	private LinkedList<Bullet> reserveList = new LinkedList<Bullet>();  // 待機弾丸リスト
	private LinkedList<Bullet> activeList = new LinkedList<Bullet>();   // 稼働弾丸リスト
	private float shotWaitTime = 0f;    // 射撃待ち時間（sec.）
	private int shotCount = 0;          // 射撃数
	private bool moving = true;			// 移動フラグ


	/// <summary>
	/// 初期化
	/// </summary>
	private void Start() {
		// 弾丸オブジェクトの事前プーリング
		GameObject prefab = Resources.Load<GameObject>("en_bullet");
		Debug.Assert(prefab.GetComponent<Bullet>() != null, "Bulletコンポーネントが存在しない");
		for (int i = 0; i < BULLET_MAX; ++i) {
			GameObject obj = Object.Instantiate(prefab);
			obj.SetActive(false);
			obj.hideFlags |= HideFlags.HideInHierarchy;
			Bullet bl = obj.GetComponent<Bullet>();
			bl.Initalize();
			this.reserveList.AddLast(bl);
		}
	}

	/// <summary>
	/// 毎フレームの更新
	/// </summary>
	private void Update() {
		float elapsedTime = Time.deltaTime;

		if (this.moving) {
			// 画面上から下に降りてくる
			this.transform.Translate(0f, 0f, -3f * elapsedTime, Space.World);
			Vector3 position = this.transform.position;
			if (position.z < 5f) {
				this.moving = false;
				position.z = 5f;
				this.transform.position = position;
			}
		} else {
			// 既存の弾丸の更新
			this.UpdateBullets(elapsedTime);
			// 新規の弾丸の更新
			this.RapidProc(elapsedTime);
		}
	}

	/// <summary>
	/// 接触時のコールバック
	/// </summary>
	/// <param name="other">接触先</param>
	private void OnTriggerEnter(Collider other) {
		// 耐久値減算
		--this.endurance;
		if (this.endurance > 0)
			return;

		// 自身のオブジェクトの破棄
		Object.Destroy(this.gameObject);
		// 爆発の生成
		Object.Instantiate(this.explosion, this.transform.position, Quaternion.identity);

		// 稼働リストがなくなるまで先頭ノードを終了させていく
		while (this.activeList.Count > 0) {
			Bullet bl = this.activeList.First.Value;
			bl.Vanish(false);
			this.reserveList.AddLast(bl);
			this.activeList.RemoveFirst();
		}
	}


	/// <summary>
	/// 連射処理
	/// </summary>
	/// <param name="elapsedTime">経過時間</param>
	private void RapidProc(float elapsedTime) {
		this.shotWaitTime -= elapsedTime;

		// TIPS: 少数点誤差調整
		if (this.shotWaitTime < 0f) {
			Bullet bl = null;
			Vector3 blPosition = this.transform.position;
			float angle = RAPID_ANGLE * shotCount;
			// 時計回り
			Quaternion rot = Quaternion.Euler(90f, angle, 0f);
			for (int i = 0; i < WAY_COUNT; ++i) {
				// 弾の取り出し
				bl = this.PickupBullet();
				// 弾切れキャンセル
				if (bl == null)
					break;
				// 射撃
				bl.Shoot(blPosition, rot, SHOT_SPEED, -this.shotWaitTime, this.activeList.Count);
				// 次の方向へ更新
				rot = WAY_ROT * rot;
			}
			// 反時計回り
			rot = Quaternion.Euler(90f, -angle, 0f);
			for (int i = 0; i < WAY_COUNT; ++i) {
				bl = this.PickupBullet();
				if (bl == null)
					break;
				bl.Shoot(blPosition, rot, SHOT_SPEED, -this.shotWaitTime, this.activeList.Count);
				rot = WAY_ROT * rot;
			}

			++this.shotCount;
			this.shotWaitTime += SHOT_SPAN;
		}
	}

	/// <summary>
	/// 待機中の弾丸の取り出し
	/// </summary>
	private Bullet PickupBullet() {
		if (this.reserveList.Count == 0) {
			Debug.LogWarning("弾切れ");
			return null;
		}

		Bullet bl = this.reserveList.First.Value;
		this.activeList.AddLast(bl);
		this.reserveList.RemoveFirst();

		return bl;
	}

	/// <summary>
	/// 弾丸の更新
	/// </summary>
	/// <param name="elapsedTime">経過時間</param>
	private void UpdateBullets(float elapsedTime) {
		// 稼働リストを全走査して終了したノードを待機リストに戻す
		LinkedListNode<Bullet> node = this.activeList.First;
		LinkedListNode<Bullet> prevNode = null;
		int index = 0;
		while (node != null) {
			Bullet bl = node.Value;
			bool active = bl.Run(index, elapsedTime);
			prevNode = node;
			node = node.Next;
			if (!active) {
				this.reserveList.AddLast(bl);
				this.activeList.Remove(prevNode);
				bl.Vanish(false);
			}
			++index;
		}
	}
}
