﻿using UnityEngine;


/// <summary>
/// 敵機
/// </summary>
public class Mob : MonoBehaviour {
	/// <summary>
	/// 移動種類
	/// </summary>
	public enum MOVE {
		DOWN,		// 画面↓
		HOMING,		// 自機追従
		TURN_LEFT,	// 左旋回
		TURN_RIGHT,	// 右旋回
	}

	private const float MOVE_SPEED = 4f;	// 移動速度（m/sec.）
	private const float HOMING_OMEGA = 60f;	// 追従角速度（degree/sec.）
	private const float TURN_OMEGA = 5f;   // 旋回角速度（degree/sec.）

	
	public GameObject explosion = null;		// 爆発オブジェクト
	public int endurance = 10;				// 耐久値
	public MOVE moveType = MOVE.DOWN;		// 移動種類設定

	private Vector3 direct = Vector3.back;  // 移動方向


	/// <summary>
	/// 毎フレームの更新
	/// </summary>
	private void Update () {
		float elapsedTime = Time.deltaTime;

		switch (this.moveType) {
			case MOVE.DOWN:
				break;
			case MOVE.HOMING:
				// 自機オブジェクトをタグから検索
				GameObject playerGo = GameObject.FindWithTag("Player");
				if (playerGo != null) {
					// 自機方向を取得
					Vector3 target = Vector3.Normalize(playerGo.transform.position - this.transform.position);
					// 現在方向と自機方向の間の角度計算（degree）
					float betweenAngle = Vector3.SignedAngle(this.direct, target, Vector3.up);
					float homingAngle = HOMING_OMEGA * elapsedTime;
					if (betweenAngle < -homingAngle) {
						// 反時計回り
						this.direct = Quaternion.AngleAxis(-homingAngle, Vector3.up) * this.direct;
					} else if (betweenAngle > homingAngle) {
						// 時計回り
						this.direct = Quaternion.AngleAxis(homingAngle, Vector3.up) * this.direct;
					} else {
						// 自機方向を補足
						this.direct = target;
					}
				}
				break;
			case MOVE.TURN_LEFT:
				this.direct = Quaternion.AngleAxis(TURN_OMEGA * elapsedTime, Vector3.up) * this.direct;
				break;
			case MOVE.TURN_RIGHT:
				this.direct = Quaternion.AngleAxis(-TURN_OMEGA * elapsedTime, Vector3.up) * this.direct;
				break;
		}
		
		// 取得した移動方向へ進行
		Vector3 position = this.transform.position + this.direct * (MOVE_SPEED * elapsedTime);
		// 移動方向へ向く回転値の計算
		Quaternion rotation = Quaternion.FromToRotation(Vector3.forward, this.direct);
		// オブジェクトへ反映
		this.transform.SetPositionAndRotation(position, rotation);
	}

	/// <summary>
	/// Renderコンポーネントが同一オブジェクトにアタッチされている場合
	/// カメラの撮影範囲から出た際に呼ばれるMonoBehaviourのコールバック
	/// </summary>
	private void OnBecameInvisible() {
		// オブジェクトの破棄
		Object.Destroy(this.gameObject);
	}

	/// <summary>
	/// 接触時のコールバック
	/// </summary>
	/// <param name="other">接触先</param>
	private void OnTriggerEnter(Collider other) {
		// 耐久値減算
		--this.endurance;
		if (this.endurance > 0)
			return;

		// 自身のオブジェクトの破棄
        Object.Destroy(this.gameObject);
		// 爆発の生成
		Object.Instantiate(this.explosion, this.transform.position, Quaternion.identity);
	}
}
