﻿using UnityEngine;
using UnityEngine.UI;
using System.Collections.Generic;


/// <summary>
/// 自機
/// </summary>
public class Fighter : MonoBehaviour {
	#region DEFINE
	private const int BULLET_MAX = 100;			// プール最大数
	private const float SHOT_SPAN = 3f / 60f;   // 射撃間隔（sec.）
	private const float SHOT_SPEED = 20f;       // 射撃速度（m/sec.）
	private const float SHOT_DELTA_ANGLE = 10f; // 射撃WAY間隔（degree）
	private const int SHOT_WAY = 3;             // 射撃WAY数
	#endregion


	#region MEMBER
	public GameObject endButton = null;			// 終了ボタン
	public GameObject explosion = null;			// 爆発オブジェクト
	public Text scoreText = null;				// スコアテキストへの参照
	
	private LinkedList<Bullet> reserveList = new LinkedList<Bullet>();	// 待機弾丸リスト
	private LinkedList<Bullet> activeList = new LinkedList<Bullet>();	// 稼働弾丸リスト
	private float shotWaitTime = 0f;			// 射撃待ち時間（sec.）
	private Vector3 startPosition = Vector3.zero;		// 入力開始時のワールド座標
	private Vector3 startInputPosition = Vector3.zero;	// 入力開始時のタッチしたワールド座標
	private int score = 0;                      // スコア値
#if !UNITY_EDITOR && !UNITY_STANDALONE // Unity EditorとPCビルドの際に有効になるシンボル
	private int touchId = 0;
#endif
	#endregion


	#region MAIN FUNCTION
	/// <summary>
	/// 初期化
	/// </summary>
	private void Start() {
		// 弾丸オブジェクトの事前プーリング
		GameObject prefab = Resources.Load<GameObject>("bullet");
		Debug.Assert(prefab.GetComponent<Bullet>() != null, "Bulletコンポーネントが存在しない");
		for (int i = 0; i < BULLET_MAX; ++i) {
			GameObject obj = Object.Instantiate(prefab);
			obj.SetActive(false);
			Bullet bl = obj.GetComponent<Bullet>();
			bl.Initalize();
			this.reserveList.AddLast(bl);
		}

		// 初回ポーズ
		Time.timeScale = 0f;

		// 自機の座標をコードから初期化する（おまけ）
		this.transform.position = new Vector3(0f, 0f, -5f);
	}

	/// <summary>
	/// 毎フレームの更新
	/// </summary>
	private void Update() {
		float elapsedTime = Time.deltaTime;

		// ポーズ中あるいは終了ボタンが表示されていたら操作不可
		if (Time.timeScale == 0f || this.endButton.activeSelf) {
			// 弾丸だけ更新
			this.UpdateBullets(elapsedTime);
			return;
		}

// Editor上とPCビルドとWebGLビルドはマウス処理
#if UNITY_EDITOR || UNITY_STANDALONE || UNITY_WEBGL
		// 左ボタン（0番）を"押した"際の処理
		if (Input.GetMouseButtonDown(0)) {
			// 現在のワールド座標を保持
			this.startPosition = this.transform.position;
			// 入力開始地点のワールド座標を保持
			this.startInputPosition = this.GetScreen2WorldPosition(Input.mousePosition);
		}
		// 左ボタン（0番）を"押している"際の処理
		if (Input.GetMouseButton(0)) {
			// 入力現在地点の座標から入力開始地点の座標を引いて移動差分の距離を計算
			// 入力開始時の座標にオフセットをかけて相対移動座標とする
			Vector3 inputPosition = this.GetScreen2WorldPosition(Input.mousePosition);
			Vector3 worldPosition = inputPosition - this.startInputPosition + this.startPosition;
			// 画面端クリッピング
			// モバイルは画面解像度が不定なのでviewport座標で対応
			Vector3 viewport = Camera.main.WorldToViewportPoint(worldPosition);
			bool clipping = false;
			float margin = 0.037f;  // 20pixの余白を持ちたい場合に1080pの場合は20/1080 = 0.037037...
			// 左右の判定
			if (viewport.x < margin || viewport.x > (1f - margin)) {
				viewport.x = Mathf.Clamp(viewport.x, margin, (1f - margin));
				clipping = true;
			}
			// 上下の判定
			margin *= Camera.main.aspect;   // 縦横のアスペクト比補正
			if (viewport.y < margin || viewport.y > (1f - margin)) {
				viewport.y = Mathf.Clamp(viewport.y, margin, (1f - margin));
				clipping = true;
			}
			if (clipping) {
				this.startInputPosition = inputPosition;
				Vector3 screenPosition = new Vector3(viewport.x * Screen.width, viewport.y * Screen.height, 0f);
				worldPosition = this.startPosition = this.GetScreen2WorldPosition(screenPosition);
			}
			// オブジェクトに反映
			this.transform.position = worldPosition;
		}
		
		// 弾丸発射処理より先に既存のものを更新
		this.UpdateBullets(elapsedTime);

		// 左ボタンを押下中に右ボタン（1番）を"押した"際の処理
		if (Input.GetMouseButtonDown(1)) {
			this.shotWaitTime = elapsedTime;
		}
		if (Input.GetMouseButton(1)) {
			this.shotWaitTime -= elapsedTime;
			// TIPS: 少数点誤差調整
			//if (this.shotWaitTime < 0f) {
			if (this.shotWaitTime < Vector3.kEpsilon) {
				// 自機の先端（前方）から発射
				Vector3 position = this.transform.position;
				position.z += 1f;
				for (int i = 0; i < SHOT_WAY; ++i) {
					if (this.reserveList.Count == 0) {
						Debug.LogWarning("弾切れ");
						break;
					}
					Bullet bl = this.reserveList.First.Value;
					// 射撃方向の計算
					Quaternion rot = Quaternion.Euler(90f, -SHOT_DELTA_ANGLE + SHOT_DELTA_ANGLE * i, 0f);
					// 弾丸が綺麗に並ぶように超過時間分だけ進める
					bl.Shoot(position, rot, SHOT_SPEED, -this.shotWaitTime, this.activeList.Count);
					this.activeList.AddLast(bl);
					this.reserveList.RemoveFirst();
				}
				this.shotWaitTime += SHOT_SPAN;
			}
		}
#else
		if (Input.touchCount > 0) {
			Touch touch = Input.GetTouch(0);
			// 左ボタン（0番）を"押した"際の処理
			if (touch.phase == TouchPhase.Began) {
				this.touchId = touch.fingerId;
				// 現在のワールド座標を保持
				this.startPosition = this.transform.position;
				// 入力開始地点のワールド座標を保持
				this.startInputPosition = this.GetScreen2WorldPosition(touch.position);
			}
			if (this.touchId == touch.fingerId) {
				// 入力現在地点の座標から入力開始地点の座標を引いて移動差分の距離を計算
				// 入力開始時の座標にオフセットをかけて相対移動座標とする
				Vector3 inputPosition = this.GetScreen2WorldPosition(touch.position);
				Vector3 worldPosition = inputPosition - this.startInputPosition + this.startPosition;
				// 画面端クリッピング
				// モバイルは画面解像度が不定なのでviewport座標で対応
				Vector3 viewport = Camera.main.WorldToViewportPoint(worldPosition);
				bool clipping = false;
				float margin = 0.037f;  // 20pixの余白を持ちたい場合に1080pの場合は20/1080 = 0.037037...
										// 左右の判定
				if (viewport.x < margin || viewport.x > (1f - margin)) {
					viewport.x = Mathf.Clamp(viewport.x, margin, (1f - margin));
					clipping = true;
				}
				// 上下の判定
				margin *= Camera.main.aspect;   // 縦横のアスペクト比補正
				if (viewport.y < margin || viewport.y > (1f - margin)) {
					viewport.y = Mathf.Clamp(viewport.y, margin, (1f - margin));
					clipping = true;
				}
				if (clipping) {
					this.startInputPosition = inputPosition;
					Vector3 screenPosition = new Vector3(viewport.x * Screen.width, viewport.y * Screen.height, 0f);
					worldPosition = this.startPosition = this.GetScreen2WorldPosition(screenPosition);
				}
				// オブジェクトに反映
				this.transform.position = worldPosition;
			}
		}
		
		// 弾丸発射処理より先に既存のものを更新
		this.UpdateBullets(elapsedTime);

		if (Input.touchCount > 1) {
			// 左ボタンを押下中に右ボタン（1番）を"押した"際の処理
			if (Input.GetTouch(1).phase == TouchPhase.Began) {
				this.shotWaitTime = elapsedTime;
			}
			this.shotWaitTime -= elapsedTime;
			// TIPS: 少数点誤差調整
			//if (this.shotWaitTime < 0f) {
			if (this.shotWaitTime < Vector3.kEpsilon) {
				// 自機の先端（前方）から発射
				Vector3 position = this.transform.position;
				position.z += 1f;
				for (int i = 0; i < SHOT_WAY; ++i) {
					if (this.reserveList.Count == 0) {
						Debug.LogWarning("弾切れ");
						break;
					}
					Bullet bl = this.reserveList.First.Value;
					// 射撃方向の計算
					Vector3 direct = Quaternion.Euler(0f, -SHOT_DELTA_ANGLE + SHOT_DELTA_ANGLE * i, 0f) * Vector3.forward;
					// 弾丸が綺麗に並ぶように超過時間分だけ進める
					bl.Shoot(position, direct, SHOT_SPEED, -this.shotWaitTime, this.activeList.Count);
					this.activeList.AddLast(bl);
					this.reserveList.RemoveFirst();
				}
				this.shotWaitTime += SHOT_SPAN;
			}
		}
#endif
	}

	/// <summary>
	/// 接触時のコールバック
	/// </summary>
	/// <param name="other">接触先</param>
	private void OnTriggerEnter(Collider other) {
		// 終了フラグON
		Object.Destroy(this.gameObject);
		// 爆発の生成
		Object.Instantiate(this.explosion, this.transform.position, Quaternion.identity);

		// 稼働リストがなくなるまで先頭ノードを終了させていく
		while (this.activeList.Count > 0) {
			Bullet bl = this.activeList.First.Value;
			bl.Vanish();
			this.reserveList.AddLast(bl);
			this.activeList.RemoveFirst();
		}

		// 終了ボタンの表示
		this.endButton.SetActive(true);
	}
	#endregion


	#region PUBLIC FUNCTION
	/// <summary>
	/// 加点
	/// </summary>
	/// <param name="point">加算する得点</param>
	public void AddScore(int point) {
		this.score += point;
		this.scoreText.text = this.score.ToString("D8");
	}
	#endregion


	#region PRIVATE FUNCTION
	/// <summary>
	/// スクリーン座標からワールド座標を求める
	/// </summary>
	/// <param name="screenPosition">入力スクリーン座標</param>
	/// <returns>入力地点のワールド座標</returns>
	private Vector3 GetScreen2WorldPosition(Vector3 screenPosition) {
		screenPosition.z = Camera.main.transform.position.y;
		return Camera.main.ScreenToWorldPoint(screenPosition);
	}

	/// <summary>
	/// 弾丸の更新
	/// </summary>
	/// <param name="elapsedTime">経過時間</param>
	private void UpdateBullets(float elapsedTime) {
		// 稼働リストを全走査して終了したノードを待機リストに戻す
		LinkedListNode<Bullet> node = this.activeList.First;
		LinkedListNode<Bullet> prevNode = null;
		int index = 0;
		while (node != null) {
			Bullet bl = node.Value;
			bool active = bl.Run(index, elapsedTime);
			prevNode = node;
			node = node.Next;
			if (!active) {
				this.reserveList.AddLast(bl);
				this.activeList.Remove(prevNode);
				bl.Vanish(false);
			}
			++index;
		}
	}
	#endregion
}
